/**********************************************************************************************
 *	@file	/env/platform.h
 *	@brief	各プラットフォームでの差異等を吸収、回避する為の定義が行われるヘッダ 
 
 *	Visual Studioの2012を最古として扱う。それより更に前へは一切の考慮が存在しない.
 *	@author	djann - kento mieno.
 *	@since	2015/10/22 22/52.
 **********************************************************************************************/
#ifndef DETOX_LIBRARY_ENVIRONMENT_ADAPTOR_DEFINITION_H_201510222252
#define DETOX_LIBRARY_ENVIRONMENT_ADAPTOR_DEFINITION_H_201510222252


// ------------------------------------------------------------
// 規定されたサイズの型をdetox名前空間へ含め統一的に扱う.
#include <cstddef>
#include <cstdint>

namespace detox {

	using std::size_t;
	using std::int8_t;
	using std::int16_t;
	using std::int32_t;
	using std::int64_t;
	using std::uint8_t;
	using std::uint16_t;
	using std::uint32_t;
	using std::uint64_t;

	using std::intptr_t;
	using std::uintptr_t;

}	// namespace detox.


#ifdef _DEBUG
	#define DETOX_DEBUG (1)
#endif	// ifdef _DEBUG.


// ============================================================
// Visual Studio向けの回避策定義.

// detoxライブラリの全てのソースはUTF8(BOM無し)のLFで作成されているが、VC++上ではwarning C4819が出る 
// #pragma warning(disable:4819)とここに書いても駄目なので可能ならばプロジェクト設定等で抑制すべきとなる.

#ifdef _MSC_VER

	// バージョン定義 - このヘッダ中で消しているので外では使えない.
	#if (_MSC_VER < 1700)
		#define UNDER_VC2012 (1)
	#else	// if (_MSC_VER < 1700).
		#define UNDER_VC2012 (0)
	#endif	// if (_MSC_VER < 1700).
	#if (_MSC_VER == 1700)
		#define VC2012 (1)
	#else	// if (_MSC_VER == 1700).
		#define VC2012 (0)
	#endif	// if (_MSC_VER == 1700).
	#if (_MSC_VER == 1800)
		#define VC2013 (1)
	#else	// if (_MSC_VER == 1800).
		#define VC2013 (0)
	#endif	// if (_MSC_VER == 1800).
	#if (_MSC_VER == 1900)
		#define VC2015 (1)
	#else	// if (_MSC_VER == 1900).
		#define VC2015 (0)
	#endif	// if (_MSC_VER == 1900).
	#if (1900 < _MSC_VER)
		#define OVER_VC2015 (1)
	#else	// if (1900 < _MSC_VER).
		#define OVER_VC2015 (0)
	#endif	// if (1900 < _MSC_VER).


	// 各バージョンに応じた定義作成.

	// Visual Studio 2015からのキーワードを以前に書く用 --------------------------------------
	#if (!VC2015 && !OVER_VC2015)
		// Clang for Windows in Visual Studioを用いていたらビルド時は省かれることになる.
		// プロジェクトに__clang__ == 1の定義を行っておく等で実際にビルドされる内容が視認可能？.
		#ifndef __clang__

			#define constexpr
			#define noexcept throw()

			#define alignas( x ) __declspec(align(x))
			#define alignof( x ) __alignof( x )

			#define thread_local __declspec(thread)

			#define __func__ __FUNCTION__

		#endif	// ifndef __clang__.
	#endif	// if (!VC2015 && !OVER_VC2015).

	// Visual Studio 2013からの対応機能を持っていないという事を表す ---------------------------
	#if (VC2012 || UNDER_VC2012)
	 	// Clang for Windows in Visual Studioを用いていたらビルド時は定義されないことになる 
		// プロジェクトに__clang__ == 1の定義を行っておく等で実際にビルドされる内容が視認可能？.
		#ifndef __clang__

			// 可変長引数テンプレート･･･.
			#define DETOX_AVAILABLE_VARIADIC_TEMPLATE (0)
			// 関数テンプレートへのデフォルトテンプレート引数･･･.
			#define DETOX_AVAILABLE_DEFAULT_ARGUMENT_FOR_TEMPLATE_FUNCTION (0)
			// 明示的型変換演算子･･･.
			#define DETOX_AVAILABLE_EXPLICIT_CONVERSION_OPERATOR (0)

		#endif	// ifndef __clang__.
	#endif	// if (VC2012 || UNDER_VC2012).

	// Visual Studio 2012より前はC++11ライブラリがほとんどない状態.
	#if (UNDER_VC2012)
		#define DETOX_AVAILABLE_AFTER_CPP11_LIBRARY (0)
	#endif	// if (UNDER_VC2012).


	// 可変長引数テンプレートに限ってはインテリセンス時のみ常に存在すると仮定する - 表示は壊れないので.
	#ifdef __INTELLISENSE__
		// ここで立っているのは0しか有り得ない.
		#ifdef DETOX_AVAILABLE_VARIADIC_TEMPLATE
			#undef DETOX_AVAILABLE_VARIADIC_TEMPLATE
		#endif	// ifdef DETOX_AVAILABLE_VARIADIC_TEMPLATE.
	#endif	// ifdef __INTELLISENSE__.

	// これらの機能を使用可能かの定義 - 上で定義されていなければこちらで1に定義する.
	#ifndef DETOX_AVAILABLE_VARIADIC_TEMPLATE
		#define DETOX_AVAILABLE_VARIADIC_TEMPLATE (1)
	#endif	// ifndef DETOX_AVAILABLE_VARIADIC_TEMPLATE.
	#ifndef DETOX_AVAILABLE_DEFAULT_ARGUMENT_FOR_TEMPLATE_FUNCTION
		#define DETOX_AVAILABLE_DEFAULT_ARGUMENT_TEMPLATE_FUNCTION (1)
	#endif	// ifndef DETOX_AVAILABLE_DEFAULT_ARGUMENT_FOR_TEMPLATE_FUNCTION.
	#ifndef DETOX_AVAILABLE_EXPLICIT_CONVERSION_OPERATOR
		#define DETOX_AVAILABLE_EXPLICIT_CONVERSION_OPERATOR (1)
	#endif	// ifndef DETOX_AVAILABLE_EXPLICIT_CONVERSION_OPERATOR.
	#ifndef DETOX_AVAILABLE_AFTER_CPP11_LIBRARY
		#define DETOX_AVAILABLE_AFTER_CPP11_LIBRARY (1)
	#endif	// ifndef DETOX_AVAILABLE_AFTER_CPP11_LIBRARY.


	// Visual Studioのバージョン定義のundef.
	#undef OVER_VC2015
	#undef VC2015
	#undef VC2013
	#undef VC2012
	#undef UNDER_VC2012


	// ========================================================================================
	// 可変長引数テンプレートが使えないVisual Studio 2012までの場合 
	// マクロ展開によるループを用い、オーバーロードで可変長引数をエミュレート実装する.
	#if !DETOX_AVAILABLE_VARIADIC_TEMPLATE

		// 引数に渡した数字が +1 されて展開されるマクロ.
		#define DETOX_PP_ADD( x ) DEPOT_PP_ADD##x
		#define DETOX_PP_ADD0 1
		#define DETOX_PP_ADD1 2
		#define DETOX_PP_ADD2 3
		#define DETOX_PP_ADD3 4
		#define DETOX_PP_ADD4 5
		#define DETOX_PP_ADD5 6
		#define DETOX_PP_ADD6 7
		#define DETOX_PP_ADD7 8
		#define DETOX_PP_ADD8 9
		#define DETOX_PP_ADD9 10

		// 引数に渡した数字が -1 されて展開されるマクロ.
		#define DETOX_PP_SUB( x ) DETOX_PP_SUB##x
		#define DETOX_PP_SUB10 9
		#define DETOX_PP_SUB9 8
		#define DETOX_PP_SUB8 7
		#define DETOX_PP_SUB7 6
		#define DETOX_PP_SUB6 5
		#define DETOX_PP_SUB5 4
		#define DETOX_PP_SUB4 3
		#define DETOX_PP_SUB3 2
		#define DETOX_PP_SUB2 1
		#define DETOX_PP_SUB1 0

		// 引数に渡した数字が0なら展開後にも0、それ以外の10までの数字は1になる条件式マクロ.
		#define DETOX_PP_TO_BOOL( n ) DETOX_PP_TO_BOOL##n
		#define DETOX_PP_TO_BOOL0  0
		#define DETOX_PP_TO_BOOL1  1
		#define DETOX_PP_TO_BOOL2  1
		#define DETOX_PP_TO_BOOL3  1
		#define DETOX_PP_TO_BOOL4  1
		#define DETOX_PP_TO_BOOL5  1
		#define DETOX_PP_TO_BOOL6  1
		#define DETOX_PP_TO_BOOL7  1
		#define DETOX_PP_TO_BOOL8  1
		#define DETOX_PP_TO_BOOL9  1
		#define DETOX_PP_TO_BOOL10 1

		// 条件式マクロを用い展開し、thenかfalse_の引数に与えたどちらかに展開するマクロ.
		#define DETOX_PP_IF( n, then, false_ ) DETOX_PP_IF##n( then, false_ )
		#define DETOX_PP_IF0( then, false_ ) false_
		#define DETOX_PP_IF1( then, false_ ) then


		// DETOX_PP_SUB_LOOP_A( number )でnumber回のループが行われる 
		// ネストした場合に同じマクロ定義があるとそこで展開が止まるので、階層毎に変える用にA, B, Cを用意している.
		#define DETOX_PP_NUM_LOOP_A( n, x ) DETOX_PP_NUM_LOOP_A##n( n, x )
		#define DETOX_PP_NUM_LOOP_A0( n, x )
		#define DETOX_PP_NUM_LOOP_A1( n, x ) DETOX_PP_NUM_LOOP_A0( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_A2( n, x ) DETOX_PP_NUM_LOOP_A1( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_A3( n, x ) DETOX_PP_NUM_LOOP_A2( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_A4( n, x ) DETOX_PP_NUM_LOOP_A3( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_A5( n, x ) DETOX_PP_NUM_LOOP_A4( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_A6( n, x ) DETOX_PP_NUM_LOOP_A5( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_A7( n, x ) DETOX_PP_NUM_LOOP_A6( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_A8( n, x ) DETOX_PP_NUM_LOOP_A7( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_A9( n, x ) DETOX_PP_NUM_LOOP_A8( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_A10( n, x ) DETOX_PP_NUM_LOOP_A9( DEPOT_PP_SUB(n), x ) x(n)

		#define DETOX_PP_NUM_LOOP_B( n, x ) DETOX_PP_NUM_LOOP_B##n( n, x )
		#define DETOX_PP_NUM_LOOP_B0( n, x )
		#define DETOX_PP_NUM_LOOP_B1( n, x ) DETOX_PP_NUM_LOOP_B0( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_B2( n, x ) DETOX_PP_NUM_LOOP_B1( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_B3( n, x ) DETOX_PP_NUM_LOOP_B2( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_B4( n, x ) DETOX_PP_NUM_LOOP_B3( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_B5( n, x ) DETOX_PP_NUM_LOOP_B4( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_B6( n, x ) DETOX_PP_NUM_LOOP_B5( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_B7( n, x ) DETOX_PP_NUM_LOOP_B6( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_B8( n, x ) DETOX_PP_NUM_LOOP_B7( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_B9( n, x ) DETOX_PP_NUM_LOOP_B8( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_B10( n, x ) DETOX_PP_NUM_LOOP_B9( DEPOT_PP_SUB(n), x ) x(n)

		#define DETOX_PP_NUM_LOOP_C( n, x ) DETOX_PP_NUM_LOOP_C##n( n, x )
		#define DETOX_PP_NUM_LOOP_C0( n, x ) 
		#define DETOX_PP_NUM_LOOP_C1( n, x )  DETOX_PP_NUM_LOOP_C0( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_C2( n, x )  DETOX_PP_NUM_LOOP_C1( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_C3( n, x )  DETOX_PP_NUM_LOOP_C2( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_C4( n, x )  DETOX_PP_NUM_LOOP_C3( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_C5( n, x )  DETOX_PP_NUM_LOOP_C4( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_C6( n, x )  DETOX_PP_NUM_LOOP_C5( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_C7( n, x )  DETOX_PP_NUM_LOOP_C6( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_C8( n, x )  DETOX_PP_NUM_LOOP_C7( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_C9( n, x )  DETOX_PP_NUM_LOOP_C8( DEPOT_PP_SUB(n), x ) x(n)
		#define DETOX_PP_NUM_LOOP_C10( n, x ) DETOX_PP_NUM_LOOP_C9( DEPOT_PP_SUB(n), x ) x(n)


		// 通常のコンマ.
		#define DETOX_PP_COMMA ,
		// DETOX_PP_NUM_LOOP_Xの使用前提で、最初の一個は存在しないコンマ
		#define DETOX_PP_NOT_FIRST_COMMA( n ) DEPOT_PP_NOT_FIRST_COMMA##n
		#define DETOX_PP_NOT_FIRST_COMMA1 
		#define DETOX_PP_NOT_FIRST_COMMA2 , 
		#define DETOX_PP_NOT_FIRST_COMMA3 , 
		#define DETOX_PP_NOT_FIRST_COMMA4 , 
		#define DETOX_PP_NOT_FIRST_COMMA5 , 
		#define DETOX_PP_NOT_FIRST_COMMA6 , 
		#define DETOX_PP_NOT_FIRST_COMMA7 , 
		#define DETOX_PP_NOT_FIRST_COMMA8 , 
		#define DETOX_PP_NOT_FIRST_COMMA9 , 
		#define DETOX_PP_NOT_FIRST_COMMA10 , 

	#endif	// if !DETOX_AVAILABLE_VARIADIC_TEMPLATE.


#endif	// ifdef _MSC_VER.


#endif	// ifndef DETOX_LIBRARY_ENVIRONMENT_ADAPTOR_DEFINITION_H_201510222252.

/* [EOF] */

