/**********************************************************************************************
 *	@file	mutant.h
 *	@brief	指定したサイズ内でどんな型も受け入れてしまう危険だが有用な機能 
 
 *	何かを作成する際に内部的に利用する事を推奨.
 *	@author	djann - kento mieno.
 *	@since	2015/11/02 23:56.
 **********************************************************************************************/

/**********************************************************************************************
					How to use for detox::mutant.

 detox::mutant can be hold any type object.

 1. include the mutant.h
	#include "detox/mutant.h"


 2. declaration.
	detox::mutant< storage-size, alignment-request(default 0x08) > m;


 3. assignment value.
	detox::mutant< 8 > m = 10;	// can be copy construct.

	// check to contain type.
	if ( m.type_of<int>() ){
		// reachable.
	}

	m = 2.5f;	// overwrite.

	if ( m.type_of<int>() ){
		// unreachable.
	} else if ( m.type_of<float>() ){
		// reachable.
	}


 4. get a contain value.
	
	float f = m.get<float>();	// type selection is your responsibility.

	// cast like.
	m = &f;	// pointer.
	std::uintptr_t ptr = m.get<std::uintptr_t>();


 5. check state.

	// has value.
	if ( m.valid() ){
		// assigned a value.
	} else {
		// no assigned.
	}

	// contain type.
	if ( m.type_of<float>() ){
		// has float value.
	}


 **********************************************************************************************/
#ifndef DETOX_TEMPLATE_LIBRARY_MUTANT_VARIABLE_TEMPLATE_H_201511022356
#define DETOX_TEMPLATE_LIBRARY_MUTANT_VARIABLE_TEMPLATE_H_201511022356


#include "./env/platform.h"

#include "./type_feature.h"


namespace detox {


/**
 *	@param	[in]	Size	使用バッファサイズ。実際のサイズはこれより大きくなる.
 *	@param	[in]	Align	要求するアライン.
 *	@note	指示無しのアライン要求はC++11標準と大体同じとなる8 
 *	@attention	アライン要求は2のべき乗とならなければならない.
 */
template <size_t Size, size_t Align = 0x08>
class mutant {
	typedef mutant<Size, Align> this_type;
	// 動的確保されてもアラインが取れるよう.
	detox::aligned_storage_has_redundancy<Size, Align> m_data;
	// 破棄される際にはデストラクタが呼ばれるが、型に応じたそれを行う為の機構をこちらに用意する.
	void (this_type::* m_deletor)();
	template <class vType>
	void call_destructor(){
		vType *ptr = reinterpret_cast<vType*>(this->m_data.get());
		ptr->~vType();
	}
public:
	// --------------------------------------------------------
	// dtor.
	~mutant(){
		// 入っているものの破棄はちゃんと行うという新しいコンセプト.
		if ( this->m_deletor ){
			(this->*m_deletor)();
			this->m_deletor = nullptr;
		}
	}

	// --------------------------------------------------------
	// ctors.
	mutant() : m_data(), m_deletor( nullptr ) {}

	// mutant同士のコピー構築は･･･一旦無しかな.
	private:
	mutant( mutant const &other );
	public:
	
	// なんらかが放り込まれての構築.
	template <class vType>
	mutant( vType const &src ) : m_data(), m_deletor( nullptr ) {
		new( this->m_data.get() ) vType( src );
		this->m_deletor = &this_type::call_destructor<vType>;
	}

	// 明示的ムーブ構築 - と言ってもコードの内容はほぼ変わらない.
	template <class vType>
	mutant( vType &&src, typename detox::enable_if<detox::is_rvalue_reference<vType&&>, void*>::type = nullptr ) : m_data(), m_deletor( nullptr ) {
		new( this->m_data.get() ) vType( detox::forward<vType>( src ) );
		this->m_deletor = &this_type::call_destructor<vType>;
	}

	// --------------------------------------------------------
	// 現在の型判別機能.
	template <class vType>
	bool type_of(){
		return this->m_deletor == &this_type::call_destructor<vType>;
	}

	// --------------------------------------------------------
	// 現在データが設定済みか否か.
	bool valid() const {
		// 設定済みならm_deletorは存在するはずである.
		//return this->m_deletor;
		// Visual Studioが出す警告抑制の為規格準拠の暗黙変換を信じないことにする.
		return this->m_deletor != nullptr;
	}

	// --------------------------------------------------------
	// コピー/ムーブ代入演算子 
	// 現在と同じ型であれば元の型の代入を、そうでなければ再構築.
	template <class vType>
	this_type const &operator = ( vType const &rhs ){
		if ( this->type_of<vType>() ){
			vType *_this = reinterpret_cast<vType*>(this->m_data.get());
			*_this = rhs;
		} else {
			this->~this_type();
			new( this ) this_type( rhs );
		}
		return *this;
	}
	// 配列が渡されたら先頭アドレスに変換しなくては.
	template <class vType, size_t N>
	this_type const &operator = ( vType const (&rhs)[N] ){
		this->~this_type();
		new( this ) this_type( const_cast<vType*>(&rhs[0]) );
		return *this;
	}

	// move.
	template <class vType>
	auto operator = ( vType &&rhs ) -> typename detox::enable_if<detox::is_rvalue_reference<vType&&>, this_type>::type const & {
		if ( this->type_of<vType>() ){
			vType *_this = reinterpret_cast<vType*>(this->m_data.get());
			*_this = detox::forward<vType>( rhs );
		} else {
			this->~this_type();
			new( this ) this_type( detox::forward<vType>( rhs ) );
		}
		return *this;
	}


	// --------------------------------------------------------
	// 特定のデータ型として取り出す.

	template <class vType>
	vType get(){
		// 検査はいらないか - union的使い方を封じないよう.
		//if ( this->type_of<vType>() )
		return *(reinterpret_cast<typename detox::remove_reference<vType>::type*>(this->m_data.get()));
	}

	// constが付く場合は中に持つ値を変更できないようconst参照を戻す.
	template <class vType>
	auto get() const ->
		typename detox::disable_if<detox::is_pointer<vType>, typename detox::remove_reference<vType>::type>::type const &
	{
		// 検査はいらないか - union的使い方を封じないよう.
		//if ( this->type_of<vType>() )
		return *(reinterpret_cast<typename detox::remove_reference<vType>::type*>(this->m_data.get()));
	}
	// ポインタの際も、軽油して中身を書き換えられなくならないように.
	template <class vType>
	auto get() const ->
		typename detox::enable_if<detox::is_pointer<vType>, typename detox::remove_pointer<vType>::type>::type const *
	{
		// 検査はいらないか - union的使い方を封じないよう.
		//if ( this->type_of<vType>() )
		return *(reinterpret_cast<vType*>(this->m_data.get()));
	}

};



}	// namespace detox.




#endif	// ifndef DETOX_TEMPLATE_LIBRARY_MUTANT_VARIABLE_TEMPLATE_H_201511022356.

/* [EOF] */

