/**********************************************************************************************
 *	@file	type_feature.h
 *	@brief	型の特徴検出等、型に対しての操作を行うライブラリ用ヘッダ.
 *	@author	djann - kento mieno.
 *	@since	2015/10/22 22:47.
 **********************************************************************************************/
#ifndef DETOX_TEMPLATE_LIBRARY_TYPE_FEATURE_H_201510222247
#define DETOX_TEMPLATE_LIBRARY_TYPE_FEATURE_H_201510222247



#include "./env/platform.h"



namespace detox {


// 基本的にdetox内部で用いて外部では使わない想定の型や関数等を定義.
namespace impl_helper {
	// detox/type_feature中でのtrue/falseを表す型 
	// メタ関数群はこのどちらかから派生することで判定結果を戻す.
	struct true_type {
		static const bool value = true;
		constexpr operator bool() const { return true; }
	};
	struct false_type {
		static const bool value = false;
		constexpr operator bool() const { return false; }
	};

	template <class vType>
	struct type_holder {
		typedef vType type;
	};


	// --------------------------------------------------------
	// アラインメント操作周りの内部使用型.

	// アライン要求のある型を2のべき乗毎に128まで.
	struct alignas(1) align1_type { char a[1]; };
	struct alignas(2) align2_type { char a[2]; };
	struct alignas(4) align4_type { char a[4]; };
	struct alignas(8) align8_type { char a[8]; };
	struct alignas(16) align16_type { char a[16]; };
	struct alignas(32) align32_type { char a[32]; };
	struct alignas(64) align64_type { char a[64]; };
	struct alignas(128) align128_type { char a[128]; };

	// 型サイズからアラインを取得する為のヘルパ.
	template <class vType>
	struct alignment_check_helper {
		char span;
		vType obj;
	};
	// 上記を用いてアライン要求を取得する.
	template <class vType>
	struct alignment_of {
		enum { value = sizeof(alignment_check_helper<vType>) - sizeof(vType) };
	};

	// -----------------------------------------------------------
	// アライン要求から、そのアラインを包含した 
	// 一番小さいalignXX_type構造体を得る.

	// どれにも合致しなかった場合 - 未定義によりコンパイルエラー.
	template <size_t Aligned, class vType, bool Condition> struct aligned_type_of;
	// マッチした場合の特殊化.
	template <size_t Aligned, class vType>
	struct aligned_type_of<Aligned, vType, true> {
		typedef vType type;
	};

	// 各種alignXX_type構造体向けの特殊化 ---------------------
	template <size_t Aligned>
	struct aligned_type_of<Aligned, align64_type, false> {
		typedef typename aligned_type_of<Aligned, align128_type, Aligned <= alignment_of<align128_type>::value>::type type;
	};
	template <size_t Aligned>
	struct aligned_type_of<Aligned, align32_type, false> {
		typedef typename aligned_type_of<Aligned, align64_type, Aligned <= alignment_of<align64_type>::value>::type type;
	};
	template <size_t Aligned>
	struct aligned_type_of<Aligned, align16_type, false> {
		typedef typename aligned_type_of<Aligned, align32_type, Aligned <= alignment_of<align32_type>::value>::type type;
	};
	template <size_t Aligned>
	struct aligned_type_of<Aligned, align8_type, false> {
		typedef typename aligned_type_of<Aligned, align16_type, Aligned <= alignment_of<align16_type>::value>::type type;
	};
	template <size_t Aligned>
	struct aligned_type_of<Aligned, align4_type, false> {
		typedef typename aligned_type_of<Aligned, align8_type, Aligned <= alignment_of<align8_type>::value>::type type;
	};
	template <size_t Aligned>
	struct aligned_type_of<Aligned, align2_type, false> {
		typedef typename aligned_type_of<Aligned, align4_type, Aligned <= alignment_of<align4_type>::value>::type type;
	};
	template <size_t Aligned>
	struct aligned_type_of<Aligned, align1_type, false> {
		typedef typename aligned_type_of<Aligned, align2_type, Aligned <= alignment_of<align2_type>::value>::type type;
	};
	

	// アライン要求とサイズ要求を両方満たしたストレージを構築する為の共用体.
	template <class AlignType, size_t Size>
	union aligned_with_size_storage_type {
		AlignType align_type;
		char size_type[ Size ];
	};


	// アライン要求とサイズ要求に合致する型を得る -------------
	template <size_t Size, size_t Align>
	struct aligned_storage_type_of {
		typedef aligned_with_size_storage_type<
			typename aligned_type_of<Align, align1_type, Align <= alignment_of<align1_type>::value>::type,
			Size
		> type;
	};

}	// namespace detox_helper.


// ============================================================
// 型選択.

// 単純な条件による型選択.
template <bool Condition, class ThenType, class FalseType>
struct type_if_v {
	typedef ThenType type;
};
template <class ThenType, class FalseType>
struct type_if_v<false, ThenType, FalseType> {
	typedef FalseType type;
};
// true_type/false_typeから派生した型が渡される事で型条件式を
template <class ConditionType, class ThenType, class FalseType>
struct type_if : public type_if_v<ConditionType::value, ThenType, FalseType> {};


// 与えた型の中から最大サイズの型を得る.
#if DETOX_AVAILABLE_VARIADIC_TEMPLATE
template <class vType, class ...vTypes>
class biggest_type_of {
	typedef typename biggest_type_of<vTypes...>::type next_type;
public:
	typedef typename type_if_v<sizeof(vType) < sizeof(next_type), next_type, vType>::type type;
};
// ラス1用特殊化.
template <class vType>
class biggest_type_of<vType> {
public:
	typedef vType type;
};
#else	// if DETOX_AVAILABLE_VARIADIC_TEMPLATE.

// プリプロセッサループでもそのままじゃ無理想なんでいっそ手書きコピペの方が良さそうである 
// 内容は上記可変長引数テンプレートの力技版なだけなので、解析はそちらを参照.
template <class vType1, class vType2 = void, class vType3 = void, class vType4 = void, class vType5 = void, class vType6 = void, class vType7 = void, class vType8 = void, class vType9 = void, class vType10 = void>
class biggest_type_of {
	typedef typename biggest_type_of< vType2, vType3, vType4, vType5, vType6, vType7, vType8, vType9, vType10, void >::type next_type;
public:
	typedef typename type_if_v< sizeof(vType1) < sizeof(next_type), next_type, vType1 >::type type;
};
template <class vType1, class vType2, class vType3, class vType4, class vType5, class vType6, class vType7, class vType8, class vType9 >
class biggest_type_of<vType1, vType2, vType3, vType4, vType5, vType6, vType7, vType8, vType9, void> {
	typedef typename biggest_type_of< vType2, vType3, vType4, vType5, vType6, vType7, vType8, vType9, void, void >::type next_type;
public:
	typedef typename type_if_v< sizeof(vType1) < sizeof(next_type), next_type, vType1 >::type type;
};
template <class vType1, class vType2, class vType3, class vType4, class vType5, class vType6, class vType7, class vType8 >
class biggest_type_of<vType1, vType2, vType3, vType4, vType5, vType6, vType7, vType8, void, void> {
	typedef typename biggest_type_of< vType2, vType3, vType4, vType5, vType6, vType7, vType8, void, void, void >::type next_type;
public:
	typedef typename type_if_v< sizeof(vType1) < sizeof(next_type), next_type, vType1 >::type type;
};
template <class vType1, class vType2, class vType3, class vType4, class vType5, class vType6, class vType7 >
class biggest_type_of<vType1, vType2, vType3, vType4, vType5, vType6, vType7, void, void, void> {
	typedef typename biggest_type_of< vType2, vType3, vType4, vType5, vType6, vType7, void, void, void, void >::type next_type;
public:
	typedef typename type_if_v< sizeof(vType1) < sizeof(next_type), next_type, vType1 >::type type;
};
template <class vType1, class vType2, class vType3, class vType4, class vType5, class vType6 >
class biggest_type_of<vType1, vType2, vType3, vType4, vType5, vType6, void, void, void, void> {
	typedef typename biggest_type_of< vType2, vType3, vType4, vType5, vType6, void, void, void, void, void >::type next_type;
public:
	typedef typename type_if_v< sizeof(vType1) < sizeof(next_type), next_type, vType1 >::type type;
};
template <class vType1, class vType2, class vType3, class vType4, class vType5 >
class biggest_type_of<vType1, vType2, vType3, vType4, vType5, void, void, void, void, void> {
	typedef typename biggest_type_of< vType2, vType3, vType4, vType5, void, void, void, void, void, void >::type next_type;
public:
	typedef typename type_if_v< sizeof(vType1) < sizeof(next_type), next_type, vType1 >::type type;
};
template <class vType1, class vType2, class vType3, class vType4 >
class biggest_type_of<vType1, vType2, vType3, vType4, void, void, void, void, void, void> {
	typedef typename biggest_type_of< vType2, vType3, vType4, void, void, void, void, void, void, void >::type next_type;
public:
	typedef typename type_if_v< sizeof(vType1) < sizeof(next_type), next_type, vType1 >::type type;
};
template <class vType1, class vType2, class vType3 >
class biggest_type_of<vType1, vType2, vType3, void, void, void, void, void, void, void> {
	typedef typename biggest_type_of< vType2, vType3, void, void, void, void, void, void, void, void >::type next_type;
public:
	typedef typename type_if_v< sizeof(vType1) < sizeof(next_type), next_type, vType1 >::type type;
};
template <class vType1, class vType2 >
class biggest_type_of<vType1, vType2, void, void, void, void, void, void, void, void> {
	typedef typename biggest_type_of< vType2, void, void, void, void, void, void, void, void, void >::type next_type;
public:
	typedef typename type_if_v< sizeof(vType1) < sizeof(next_type), next_type, vType1 >::type type;
};
template <class vType>
class biggest_type_of<vType, void, void, void, void, void, void, void, void, void> {
public:
	typedef vType type;
};

#endif	// if DETOX_AVAILABLE_VARIADIC_TEMPLATE.


// 条件式がtrueの場合にのみ内部型typeを持つ、つまりコンパイルが通ることになる 
// 言わずと知れたenable_ifである.
template <bool Condition, class Type = void>
struct enable_if_v {};
template <class Type>
struct enable_if_v<true, Type> { typedef Type type; };
template <>
struct enable_if_v<true, void> { typedef void *&type; };

template <class ConditionType, class Type = void>
struct enable_if : public enable_if_v<ConditionType::value, Type> {};

// enable_ifの逆であり、条件式がfalseの場合にのみ内部型typeを持ちコンパイルを通す.
template <bool Condition, class Type>
struct disable_if_v : public enable_if_v<!Condition, Type> {};
template <class ConditionType, class Type>
struct disable_if : public enable_if_v<!ConditionType::value, Type> {};

// 言わずと知れたenabler idiom.
extern void *enabler;




// ============================================================
// 型の性質を判別.

// その型はポインタか？.
template <class vType>
struct is_pointer : public impl_helper::false_type {};
template <class vType>
struct is_pointer<vType*> : public impl_helper::true_type {};
template <class vType>
struct is_pointer<vType* const> : public impl_helper::true_type {};
template <class vType>
struct is_pointer<vType* volatile> : public impl_helper::true_type {};

// その型は左辺値参照か？.
template <class vType>
struct is_lvalue_reference : public impl_helper::false_type {};
template <class vType>
struct is_lvalue_reference<vType&> : public impl_helper::true_type {};

// その型は右辺値参照か？.
template <class vType>
struct is_rvalue_reference : public impl_helper::false_type {};
template <class vType>
struct is_rvalue_reference<vType&&> : public impl_helper::true_type {};

// 参照か？.
template <class vType>
struct is_reference
	: public detox::type_if_v<detox::is_lvalue_reference<vType>::value || detox::is_rvalue_reference<vType>::value, impl_helper::true_type, impl_helper::false_type>::type
{
};

// voidか？.
template <class vType>
struct is_void : public impl_helper::false_type {};
template <>
struct is_void<void> : public impl_helper::true_type {};

// 二つは同じ型か？.
template <class vType1, class vType2>
struct is_same : public impl_helper::false_type {};
template <class vType1>
struct is_same<vType1, vType1> : public impl_helper::true_type {};

// 配列型～とか滅多にないだろうけど.
template <class vType>
struct is_array : public impl_helper::false_type {};
template <class vType, size_t N>
struct is_array<vType[N]> : public impl_helper::true_type {};
template <class vType>
struct is_array<vType[]> : public impl_helper::true_type {};




// ============================================================
// 型から一部特製を抜いたり追加する操作.

// ポインタを取り除いた型を得る -------------------------------
template <class vType>
struct remove_pointer : public impl_helper::type_holder<vType> {};
template <class vType>
struct remove_pointer<vType*> : public impl_helper::type_holder<vType> {};
template <class vType>
struct remove_pointer<vType *const> : public impl_helper::type_holder<vType> {};
template <class vType>
struct remove_pointer<vType *volatile> : public impl_helper::type_holder<vType> {};
template <class vType>
struct remove_pointer<vType *volatile const> : public impl_helper::type_holder<vType> {};

// 参照を取り除いた型を得る -----------------------------------
template <class vType>
struct remove_reference : public impl_helper::type_holder<vType> {};
template <class vType>
struct remove_reference<vType&> : public impl_helper::type_holder<vType> {};
template <class vType>
struct remove_reference<vType&&> : public impl_helper::type_holder<vType> {};


// ポインタを追加した型を得る ---------------------------------
template <class vType>
struct add_pointer : public impl_helper::type_holder<typename detox::remove_reference<vType>::type*> {};

// 参照を追加した型を得る -------------------------------------
template <class vType>
struct add_lvalue_reference : public impl_helper::type_holder<typename detox::remove_reference<vType>::type&> {};
template <class vType>
struct add_rvalue_reference : public impl_helper::type_holder<typename detox::remove_reference<vType>::type&&> {};




// ============================================================
// 型から情報を取り出す.

// その型のアライン要求を得る.
template <class vType>
struct alignment_of : public impl_helper::alignment_of<vType> {};


/**
 *	@brief C++03でも使用可なアライン指定のストレージ確保テンプレートクラス 
 *	@attention C++14現在アライン要求は、自動ストレージや静的ストレージに構築された場合にのみ満たされる 

 *	aligned_storageも例外ではなく動的ストレージに構築される場合は満たされる保証がない。 
 *	動的ストレージにアライン要求が必要な場合はaligned_storage_has_redundancyの使用を推奨.
 */
template <size_t Size, size_t Align>
class aligned_storage {
	typedef typename impl_helper::aligned_storage_type_of<Size, Align>::type storage_type;
	storage_type m_storage;
public:
	// ただストレージを持つだけなので構築等の機能は一切無い.
	void *get() const noexcept { return &this->m_storage; }
};

// 型からaligned_storageを導出するインターフェース。使用方法はaligned_storageと同一である.
template <class vType>
class aligned_storage_from_type : public aligned_storage<sizeof(vType), impl_helper::alignment_of<vType>::value> {};


/**
 *	@brief	動的ストレージへ確保する際にもgetで取得可能なアドレスがアライン要求を満たすaligned_storage.
 */
template <size_t Size, size_t Align>
class aligned_storage_has_redundancy {
	uint8_t m_storage[ Size + Align ];
	void *m_ptr;
	static void *normalize( void *ptr ){
		auto addr = reinterpret_cast<uintptr_t>(ptr);
		// ぴったり？.
		decltype(addr) rem = addr % Align;
		if ( rem ){
			addr += Align - rem;
		}
		return reinterpret_cast<void*>(addr);
	}
public:
	/**	@brief	setting 'm_ptr'. */
	aligned_storage_has_redundancy() : m_ptr( normalize( m_storage ) ){}
	void *get() const noexcept { return m_ptr; }
};

// aligned_storageと同様、型からaligned_storage_has_redundancyを導出する.
template <class vType>
class aligned_storage_has_redundancy_from_type : public detox::aligned_storage_has_redundancy<sizeof(vType), impl_helper::alignment_of<vType>::value> {};




// has_XXXメタ関数を作成する為のヘルパマクロ 
// 特定のメンバやデータメンバを持った型か否かを判定し 
// detox::impl_helper::true_type or false_typeから派生する 
// そのままdetox::enable_if等で使用する事が出来る他valueメンバを参照し 
// 標準ライブラリの type_traits へ渡す事も可能.
#define DETOX_SFINAE_HAS_MEMBER_HELPER( Name, Type, Args ) \
	namespace impl_helper {\
		template <class vType, Type (vType::*)Args>\
		struct has_ ## Name ## _sfinae_helper { typedef void type; };\
	}	// namespace impl_helper.

#define DETOX_MAKE_HAS_XXX_MEMBER_FUNCTION( Name, RetType, Args ) \
	namespace detox {\
		DETOX_SFINAE_HAS_MEMBER_HELPER( Name, RetType, Args )\
		template <class vType, class SigType = void> struct has_ ## Name : public detox::impl_helper::false_type {};\
		template <class vType>\
		struct has_ ## Name<vType, typename impl_helper::has_ ## Name ## _sfinae_helper<vType, &vType::Name>::type> : public detox::impl_helper::true_type {};\
	}	// namespace detox.

// Argsが無くなれば実はほぼ同じ構文.
#define DETOX_MAKE_HAS_XXX_DATA_MEMBER( Name, Type ) DETOX_MAKE_HAS_XXX_MEMBER_FUNCTION( Name, Type, )


}	// namespace detox.


#endif	// ifndef DETOX_TEMPLATE_LIBRARY_TYPE_FEATURE_H_201510222247.

/* [EOF] */

